@@ -94,6 +94,11 @@ static bool dr_mask_is_ttl_set(struct dr_match_spec *spec)
 	return spec->ip_ttl_hoplimit;
 }
 
+static bool dr_mask_is_ipv4_ihl_set(struct dr_match_spec *spec)
+{
+	return spec->ipv4_ihl;
+}
+
 #define DR_MASK_IS_L2_DST(_spec, _misc, _inner_outer) (_spec.first_vid || \
 	(_spec).first_cfi || (_spec).first_prio || (_spec).cvlan_tag || \
 	(_spec).svlan_tag || (_spec).dmac_47_16 || (_spec).dmac_15_0 || \
@@ -199,7 +204,7 @@ dr_mask_is_tnl_geneve(struct dr_match_param *mask,
 
 static bool dr_mask_is_tnl_gtpu_set(struct dr_match_misc3 *misc3)
 {
-	return misc3->gtpu_flags || misc3->gtpu_msg_type || misc3->gtpu_teid;
+	return misc3->gtpu_msg_flags || misc3->gtpu_msg_type || misc3->gtpu_teid;
 }
 
 static bool dr_matcher_supp_tnl_gtpu(struct dr_devx_caps *caps)
@@ -214,6 +219,92 @@ static bool dr_mask_is_tnl_gtpu(struct dr_match_param *mask,
 	       dr_matcher_supp_tnl_gtpu(&dmn->info.caps);
 }
 
+static int dr_matcher_supp_tnl_gtpu_dw_0(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_GTPU_DW_0_ENABLED;
+}
+
+static bool dr_mask_is_tnl_gtpu_dw_0(struct dr_match_param *mask,
+				     struct mlx5dv_dr_domain *dmn)
+{
+	return mask->misc3.gtpu_dw_0 &&
+	       dr_matcher_supp_tnl_gtpu_dw_0(&dmn->info.caps);
+}
+
+static int dr_matcher_supp_tnl_gtpu_teid(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_GTPU_TEID_ENABLED;
+}
+
+static bool dr_mask_is_tnl_gtpu_teid(struct dr_match_param *mask,
+				     struct mlx5dv_dr_domain *dmn)
+{
+	return mask->misc3.gtpu_teid &&
+	       dr_matcher_supp_tnl_gtpu_teid(&dmn->info.caps);
+}
+
+static int dr_matcher_supp_tnl_gtpu_dw_2(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_GTPU_DW_2_ENABLED;
+}
+
+static bool dr_mask_is_tnl_gtpu_dw_2(struct dr_match_param *mask,
+				     struct mlx5dv_dr_domain *dmn)
+{
+	return mask->misc3.gtpu_dw_2 &&
+	       dr_matcher_supp_tnl_gtpu_dw_2(&dmn->info.caps);
+}
+
+static int dr_matcher_supp_tnl_gtpu_first_ext(struct dr_devx_caps *caps)
+{
+	return caps->flex_protocols & MLX5_FLEX_PARSER_GTPU_FIRST_EXT_DW_0_ENABLED;
+}
+
+static bool dr_mask_is_tnl_gtpu_first_ext(struct dr_match_param *mask,
+					  struct mlx5dv_dr_domain *dmn)
+{
+	return mask->misc3.gtpu_first_ext_dw_0 &&
+	       dr_matcher_supp_tnl_gtpu_first_ext(&dmn->info.caps);
+}
+
+static bool dr_mask_is_tnl_gtpu_flex_parser_0(struct dr_match_param *mask,
+					      struct mlx5dv_dr_domain *dmn)
+{
+	struct dr_devx_caps *caps = &dmn->info.caps;
+
+	return ((caps->flex_parser_id_gtpu_dw_0 <= DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_dw_0(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_teid <= DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_teid(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_dw_2 <= DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_dw_2(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_first_ext_dw_0 <= DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_first_ext(mask, dmn));
+}
+
+static bool dr_mask_is_tnl_gtpu_flex_parser_1(struct dr_match_param *mask,
+					      struct mlx5dv_dr_domain *dmn)
+{
+	struct dr_devx_caps *caps = &dmn->info.caps;
+
+	return ((caps->flex_parser_id_gtpu_dw_0 > DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_dw_0(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_teid > DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_teid(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_dw_2 > DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_dw_2(mask, dmn)) ||
+	       ((caps->flex_parser_id_gtpu_first_ext_dw_0 > DR_STE_MAX_FLEX_0_ID) &&
+		dr_mask_is_tnl_gtpu_first_ext(mask, dmn));
+}
+
+static bool dr_mask_is_tnl_gtpu_any(struct dr_match_param *mask,
+				    struct mlx5dv_dr_domain *dmn)
+{
+	return dr_mask_is_tnl_gtpu_flex_parser_0(mask, dmn) ||
+	       dr_mask_is_tnl_gtpu_flex_parser_1(mask, dmn) ||
+	       dr_mask_is_tnl_gtpu(mask, dmn);
+}
+
 static inline int dr_matcher_supp_icmp_v4(struct dr_devx_caps *caps)
 {
 	return (caps->sw_format_ver == MLX5_HW_CONNECTX_6DX) ||
@@ -301,6 +392,11 @@ static bool dr_mask_is_flex_parser_4_7_set(struct dr_match_misc4 *misc4)
 		dr_mask_is_flex_parser_id_4_7_set(misc4->prog_sample_field_id_3));
 }
 
+static bool dr_mask_is_tunnel_header_0_1_set(struct dr_match_misc5 *misc5)
+{
+	return misc5->tunnel_header_0 || misc5->tunnel_header_1;
+}
+
 static int dr_matcher_supp_tnl_mpls_over_gre(struct dr_devx_caps *caps)
 {
 	return caps->flex_protocols & MLX5_FLEX_PARSER_MPLS_OVER_GRE_ENABLED;
@@ -325,6 +421,239 @@ static bool dr_mask_is_tnl_mpls_over_udp(struct dr_match_param *mask,
 	       dr_matcher_supp_tnl_mpls_over_udp(&dmn->info.caps);
 }
 
+static bool dr_matcher_is_mask_consumed(struct dr_match_param *mask)
+{
+	int i;
+
+	for (i = 0; i < sizeof(struct dr_match_param); i++)
+		if (((uint8_t *)mask)[i] != 0)
+			return false;
+
+	return true;
+}
+
+static void dr_matcher_copy_mask(struct dr_match_param *dst_mask,
+				 struct dr_match_param *src_mask,
+				 uint8_t match_criteria)
+{
+	if (match_criteria & DR_MATCHER_CRITERIA_OUTER)
+		dst_mask->outer = src_mask->outer;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC)
+		dst_mask->misc = src_mask->misc;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_INNER)
+		dst_mask->inner = src_mask->inner;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC2)
+		dst_mask->misc2 = src_mask->misc2;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC3)
+		dst_mask->misc3 = src_mask->misc3;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC4)
+		dst_mask->misc4 = src_mask->misc4;
+
+	if (match_criteria & DR_MATCHER_CRITERIA_MISC5)
+		dst_mask->misc5 = src_mask->misc5;
+}
+
+static void dr_matcher_destroy_definer_objs(struct dr_ste_build *sb,
+					    uint8_t idx)
+{
+	int i;
+
+	for (i = 0; i < idx; i++) {
+		mlx5dv_devx_obj_destroy(sb[i].definer_obj);
+		sb[i].lu_type = 0;
+		sb[i].htbl_type = 0;
+		sb[i].definer_obj = NULL;
+	}
+}
+
+static int dr_matcher_create_definer_objs(struct ibv_context *ctx,
+					  struct dr_ste_build *sb,
+					  uint8_t idx)
+{
+	struct mlx5dv_devx_obj *devx_obj;
+	int i;
+
+	for (i = 0; i < idx; i++) {
+		devx_obj = dr_devx_create_definer(ctx, sb[i].format_id, sb[i].match);
+		if (!devx_obj)
+			goto cleanup;
+
+		/* The lu_type combines the definer and the entry type */
+		sb[i].lu_type |= devx_obj->object_id;
+		sb[i].htbl_type = DR_STE_HTBL_TYPE_MATCH;
+		sb[i].definer_obj = devx_obj;
+	}
+
+	return 0;
+
+cleanup:
+	dr_matcher_destroy_definer_objs(sb, i);
+	return errno;
+}
+
+static void dr_matcher_clear_definers_builders(struct dr_matcher_rx_tx *nic_matcher)
+{
+	struct dr_ste_build *sb = nic_matcher->ste_builder;
+	int i;
+
+	for (i = 0; i < nic_matcher->num_of_builders; i++)
+		memset(&sb[i], 0, sizeof(*sb));
+
+	nic_matcher->num_of_builders = 0;
+}
+
+static int dr_matcher_set_definer_builders(struct mlx5dv_dr_matcher *matcher,
+					   struct dr_matcher_rx_tx *nic_matcher)
+{
+	struct dr_domain_rx_tx *nic_dmn = nic_matcher->nic_tbl->nic_dmn;
+	struct dr_ste_build *sb = nic_matcher->ste_builder;
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	bool rx = nic_dmn->type == DR_DOMAIN_NIC_TYPE_RX;
+	struct dr_devx_caps *caps = &dmn->info.caps;
+	struct dr_ste_ctx *ste_ctx = dmn->ste_ctx;
+	struct dr_match_param mask = {};
+	bool src_ipv6, dst_ipv6;
+	uint8_t idx = 0;
+	uint8_t ipv;
+	int ret;
+
+	ipv = matcher->mask.outer.ip_version;
+	src_ipv6 = dr_mask_is_src_addr_set(&matcher->mask.outer);
+	dst_ipv6 = dr_mask_is_dst_addr_set(&matcher->mask.outer);
+
+
+	if (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_0)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def0(ste_ctx, &sb[idx++], &mask, caps, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (dmn->info.caps.definer_format_sup & (1 << DR_MATCHER_DEFINER_2)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def2(ste_ctx, &sb[idx++], &mask, caps, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_16)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def16(ste_ctx, &sb[idx++], &mask, caps, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_22)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def22(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_24)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def24(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_25)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def25(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if ((ipv == DR_MASK_IP_VERSION_IPV6 && src_ipv6) &&
+	    (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_6)) &&
+	    (caps->definer_format_sup & (1 << DR_MATCHER_DEFINER_26))) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def26(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dst_ipv6)
+			ret = dr_ste_build_def6(ste_ctx, &sb[idx++], &mask, false, rx);
+
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(&sb[0], 0, sizeof(*sb));
+		memset(&sb[1], 0, sizeof(*sb));
+		idx = 0;
+	}
+
+	if (dmn->info.caps.definer_format_sup & (1 << DR_MATCHER_DEFINER_28)) {
+		dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+		ret = dr_ste_build_def28(ste_ctx, &sb[idx++], &mask, false, rx);
+		if (!ret && dr_matcher_is_mask_consumed(&mask))
+			goto done;
+
+		memset(sb, 0, sizeof(struct dr_ste_build));
+		idx = 0;
+	}
+
+	return ENOTSUP;
+
+done:
+	nic_matcher->num_of_builders = idx;
+	return 0;
+}
+
+static int dr_matcher_set_large_ste_builders(struct mlx5dv_dr_matcher *matcher,
+					     struct dr_matcher_rx_tx *nic_matcher)
+{
+	struct mlx5dv_dr_domain *dmn = matcher->tbl->dmn;
+	int ret;
+
+	if (dmn->info.caps.sw_format_ver != MLX5_HW_CONNECTX_6DX ||
+	    !dmn->info.caps.definer_format_sup)
+		return ENOTSUP;
+
+	ret = dr_matcher_set_definer_builders(matcher, nic_matcher);
+	if (ret)
+		return ret;
+
+	ret = dr_matcher_create_definer_objs(dmn->ctx,
+					     nic_matcher->ste_builder,
+					     nic_matcher->num_of_builders);
+	if (ret)
+		goto clear_definers_builders;
+
+	return 0;
+
+clear_definers_builders:
+	dr_matcher_clear_definers_builders(nic_matcher);
+	return ret;
+}
+
+static void dr_matcher_clear_ste_builders(struct dr_matcher_rx_tx *nic_matcher)
+{
+	if (nic_matcher->ste_builder->htbl_type == DR_STE_HTBL_TYPE_MATCH)
+		dr_matcher_destroy_definer_objs(nic_matcher->ste_builder,
+						nic_matcher->num_of_builders);
+}
+
 static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 				       struct dr_matcher_rx_tx *nic_matcher)
 {
@@ -337,37 +666,25 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 	bool inner, rx;
 	uint8_t ipv;
 	int idx = 0;
-	int ret, i;
-
-	rx = nic_dmn->ste_type == DR_STE_TYPE_RX;
-
-	/* Create a temporary mask to track and clear used mask fields */
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_OUTER)
-		mask.outer = matcher->mask.outer;
-
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC)
-		mask.misc = matcher->mask.misc;
-
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_INNER)
-		mask.inner = matcher->mask.inner;
-
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC2)
-		mask.misc2 = matcher->mask.misc2;
-
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC3)
-		mask.misc3 = matcher->mask.misc3;
-
-	if (matcher->match_criteria & DR_MATCHER_CRITERIA_MISC4)
-		mask.misc4 = matcher->mask.misc4;
+	int ret;
 
 	ret = dr_ste_build_pre_check(dmn, matcher->match_criteria,
 				     &matcher->mask, NULL);
 	if (ret)
 		return ret;
 
+	/* Use a large definers for matching if possible */
+	ret = dr_matcher_set_large_ste_builders(matcher, nic_matcher);
+	if (!ret)
+		return 0;
+
+	/* Create a temporary mask to track and clear used mask fields */
+	dr_matcher_copy_mask(&mask, &matcher->mask, matcher->match_criteria);
+
 	/* Optimize RX pipe by reducing source port match, since
-	 * the FDB RX part is conneted only to the wire.
+	 * the FDB RX part is connected only to the wire.
 	 */
+	rx = nic_dmn->type == DR_DOMAIN_NIC_TYPE_RX;
 	if (dmn->type == MLX5DV_DR_DOMAIN_TYPE_FDB &&
 	    rx && mask.misc.source_port) {
 		mask.misc.source_port = 0;
@@ -378,7 +695,8 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 	if (matcher->match_criteria & (DR_MATCHER_CRITERIA_OUTER |
 				       DR_MATCHER_CRITERIA_MISC |
 				       DR_MATCHER_CRITERIA_MISC2 |
-				       DR_MATCHER_CRITERIA_MISC3)) {
+				       DR_MATCHER_CRITERIA_MISC3 |
+				       DR_MATCHER_CRITERIA_MISC5)) {
 		inner = false;
 		ipv = mask.outer.ip_version;
 
@@ -415,7 +733,8 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 						&mask, inner, rx);
 
 		if (ipv == 4) {
-			if (dr_mask_is_ttl_set(&mask.outer))
+			if (dr_mask_is_ttl_set(&mask.outer) ||
+			    dr_mask_is_ipv4_ihl_set(&mask.outer))
 				dr_ste_build_eth_l3_ipv4_misc(ste_ctx, &sb[idx++],
 							      &mask, inner, rx);
 
@@ -449,9 +768,23 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 				dr_ste_build_tnl_geneve_tlv_opt(ste_ctx, &sb[idx++],
 								&mask, &dmn->info.caps,
 								inner, rx);
-		} else if (dr_mask_is_tnl_gtpu(&mask, dmn)) {
-			dr_ste_build_tnl_gtpu(ste_ctx, &sb[idx++],
-					      &mask, inner, rx);
+		} else if (dr_mask_is_tnl_gtpu_any(&mask, dmn)) {
+			if (dr_mask_is_tnl_gtpu_flex_parser_0(&mask, dmn))
+				dr_ste_build_tnl_gtpu_flex_parser_0(ste_ctx, &sb[idx++],
+								    &mask, &dmn->info.caps,
+								    inner, rx);
+
+			if (dr_mask_is_tnl_gtpu_flex_parser_1(&mask, dmn))
+				dr_ste_build_tnl_gtpu_flex_parser_1(ste_ctx, &sb[idx++],
+								    &mask, &dmn->info.caps,
+								    inner, rx);
+
+			if (dr_mask_is_tnl_gtpu(&mask, dmn))
+				dr_ste_build_tnl_gtpu(ste_ctx, &sb[idx++],
+						      &mask, inner, rx);
+		} else if (dr_mask_is_tunnel_header_0_1_set(&mask.misc5)) {
+			dr_ste_build_tunnel_header_0_1(ste_ctx, &sb[idx++],
+						       &mask, false, rx);
 		}
 
 		if (DR_MASK_IS_ETH_L4_MISC_SET(mask.misc3, outer))
@@ -506,7 +839,8 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 						&mask, inner, rx);
 
 		if (ipv == 4) {
-			if (dr_mask_is_ttl_set(&mask.inner))
+			if (dr_mask_is_ttl_set(&mask.inner) ||
+			    dr_mask_is_ipv4_ihl_set(&mask.inner))
 				dr_ste_build_eth_l3_ipv4_misc(ste_ctx, &sb[idx++],
 							      &mask, inner, rx);
 
@@ -566,16 +900,15 @@ static int dr_matcher_set_ste_builders(struct mlx5dv_dr_matcher *matcher,
 		return errno;
 	}
 
+	/* Check that all mask fields were consumed */
+	if (!dr_matcher_is_mask_consumed(&mask)) {
+		dr_dbg(dmn, "Mask contains unsupported parameters\n");
+		errno = EOPNOTSUPP;
+		return errno;
+	}
+
 	nic_matcher->num_of_builders = idx;
 
-	/* Check that all mask fields were consumed */
-	for (i = 0; i < sizeof(struct dr_match_param); i++) {
-		if (((uint8_t *)&mask)[i] != 0) {
-			dr_dbg(dmn, "Mask contains unsupported parameters\n");
-			errno = EOPNOTSUPP;
-			return errno;
-		}
-	}
 	return 0;
 }
 
@@ -699,6 +1032,7 @@ static int dr_matcher_add_to_tbl(struct mlx5dv_dr_matcher *matcher)
 
 static void dr_matcher_uninit_nic(struct dr_matcher_rx_tx *nic_matcher)
 {
+	dr_matcher_clear_ste_builders(nic_matcher);
 	dr_htbl_put(nic_matcher->s_htbl);
 	dr_htbl_put(nic_matcher->e_anchor);
 }
@@ -751,15 +1085,17 @@ static int dr_matcher_init_nic(struct mlx5dv_dr_matcher *matcher,
 
 	nic_matcher->e_anchor = dr_ste_htbl_alloc(dmn->ste_icm_pool,
 						  DR_CHUNK_SIZE_1,
+						  DR_STE_HTBL_TYPE_LEGACY,
 						  DR_STE_LU_TYPE_DONT_CARE,
 						  0);
 	if (!nic_matcher->e_anchor)
-		return errno;
+		goto clear_ste_builders;
 
 	nic_matcher->s_htbl = dr_ste_htbl_alloc(dmn->ste_icm_pool,
 						DR_CHUNK_SIZE_1,
-						nic_matcher->ste_builder[0].lu_type,
-						nic_matcher->ste_builder[0].byte_mask);
+						nic_matcher->ste_builder->htbl_type,
+						nic_matcher->ste_builder->lu_type,
+						nic_matcher->ste_builder->byte_mask);
 	if (!nic_matcher->s_htbl)
 		goto free_e_htbl;
 
@@ -771,6 +1107,8 @@ static int dr_matcher_init_nic(struct mlx5dv_dr_matcher *matcher,
 
 free_e_htbl:
 	dr_ste_htbl_free(nic_matcher->e_anchor);
+clear_ste_builders:
+	dr_matcher_clear_ste_builders(nic_matcher);
 	return errno;
 }
 
