@@ -34,6 +34,12 @@
 
 #define DR_STE_ENABLE_FLOW_TAG (1 << 31)
 
+enum dr_ste_v0_entry_type {
+	DR_STE_TYPE_TX		= 1,
+	DR_STE_TYPE_RX		= 2,
+	DR_STE_TYPE_MODIFY_PKT	= 6,
+};
+
 enum dr_ste_v0_action_tunl {
 	DR_STE_TUNL_ACTION_NONE		= 0,
 	DR_STE_TUNL_ACTION_ENABLE	= 1,
@@ -100,6 +106,7 @@ enum dr_ste_v0_lu_type {
 	DR_STE_V0_LU_TYPE_GENERAL_PURPOSE		= 0x18,
 	DR_STE_V0_LU_TYPE_STEERING_REGISTERS_0		= 0x2f,
 	DR_STE_V0_LU_TYPE_STEERING_REGISTERS_1		= 0x30,
+	DR_STE_V0_LU_TYPE_TUNNEL_HEADER			= 0x34,
 	DR_STE_V0_LU_TYPE_DONT_CARE			= DR_STE_LU_TYPE_DONT_CARE,
 };
 
@@ -313,8 +320,9 @@ static void dr_ste_v0_set_hit_addr(uint8_t *hw_ste_p, uint64_t icm_addr, uint32_
 	DR_STE_SET(general, hw_ste_p, next_table_base_31_5_size, index);
 }
 
-static void dr_ste_v0_init(uint8_t *hw_ste_p, uint16_t lu_type,
-			   uint8_t entry_type, uint16_t gvmi)
+static void dr_ste_v0_init_full(uint8_t *hw_ste_p, uint16_t lu_type,
+				enum dr_ste_v0_entry_type entry_type,
+				uint16_t gvmi)
 {
 	dr_ste_v0_set_entry_type(hw_ste_p, entry_type);
 	dr_ste_v0_set_lu_type(hw_ste_p, lu_type);
@@ -325,6 +333,35 @@ static void dr_ste_v0_init(uint8_t *hw_ste_p, uint16_t lu_type,
 	DR_STE_SET(rx_steering_mult, hw_ste_p, miss_address_63_48, gvmi);
 }
 
+static void dr_ste_v0_init(uint8_t *hw_ste_p, uint16_t lu_type,
+			   bool is_rx, uint16_t gvmi)
+{
+	enum dr_ste_v0_entry_type entry_type;
+
+	entry_type = is_rx ? DR_STE_TYPE_RX : DR_STE_TYPE_TX;
+
+	dr_ste_v0_init_full(hw_ste_p, lu_type, entry_type, gvmi);
+}
+
+static void dr_ste_v0_set_ctrl_always_hit_htbl(uint8_t *hw_ste_p,
+					       uint16_t byte_mask,
+					       uint16_t lu_type,
+					       uint64_t icm_addr,
+					       uint32_t num_of_entries,
+					       uint16_t gvmi)
+{
+	dr_ste_v0_set_next_lu_type(hw_ste_p, lu_type);
+	dr_ste_v0_set_hit_addr(hw_ste_p, icm_addr, num_of_entries);
+	dr_ste_v0_set_byte_mask(hw_ste_p, byte_mask);
+}
+
+static void dr_ste_v0_set_ctrl_always_miss(uint8_t *hw_ste_p, uint64_t miss_addr,
+					   uint16_t gvmi)
+{
+	dr_ste_v0_set_next_lu_type(hw_ste_p, DR_STE_LU_TYPE_DONT_CARE);
+	dr_ste_v0_set_miss_addr(hw_ste_p, miss_addr);
+}
+
 static void dr_ste_v0_set_rx_flow_tag(uint8_t *hw_ste_p, uint32_t flow_tag)
 {
 	DR_STE_SET(rx_steering_mult, hw_ste_p, qp_list_pointer,
@@ -398,12 +435,12 @@ static void dr_ste_v0_set_rewrite_actions(uint8_t *hw_ste_p,
 
 static inline void dr_ste_v0_arr_init_next(uint8_t **last_ste,
 					   uint32_t *added_stes,
-					   enum dr_ste_entry_type entry_type,
+					   enum dr_ste_v0_entry_type entry_type,
 					   uint16_t gvmi)
 {
 	(*added_stes)++;
 	*last_ste += DR_STE_SIZE;
-	dr_ste_v0_init(*last_ste, DR_STE_LU_TYPE_DONT_CARE, entry_type, gvmi);
+	dr_ste_v0_init_full(*last_ste, DR_STE_LU_TYPE_DONT_CARE, entry_type, gvmi);
 }
 
 static void dr_ste_v0_set_actions_tx(uint8_t *action_type_set,
@@ -671,6 +708,25 @@ dr_ste_v0_set_action_decap_l3_list(void *data, uint32_t data_sz,
 	return 0;
 }
 
+static const struct dr_ste_action_modify_field
+*dr_ste_v0_get_action_hw_field(uint16_t sw_field, struct dr_devx_caps *caps)
+{
+	const struct dr_ste_action_modify_field *hw_field;
+
+	if (sw_field >= ARRAY_SIZE(dr_ste_v0_action_modify_field_arr))
+		goto not_found;
+
+	hw_field = &dr_ste_v0_action_modify_field_arr[sw_field];
+	if (!hw_field->end && !hw_field->start)
+		goto not_found;
+
+	return hw_field;
+
+not_found:
+	errno = EINVAL;
+	return NULL;
+}
+
 static void dr_ste_v0_build_eth_l2_src_dst_bit_mask(struct dr_match_param *value,
 						    bool inner, uint8_t *bit_mask)
 {
@@ -1101,6 +1157,7 @@ static int dr_ste_v0_build_eth_l3_ipv4_misc_tag(struct dr_match_param *value,
 	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
 
 	DR_STE_SET_TAG(eth_l3_ipv4_misc, tag, time_to_live, spec, ip_ttl_hoplimit);
+	DR_STE_SET_TAG(eth_l3_ipv4_misc, tag, ihl, spec, ipv4_ihl);
 
 	return 0;
 }
@@ -1239,9 +1296,9 @@ static void dr_ste_v0_build_tnl_mpls_over_udp_init(struct dr_ste_build *sb,
 	/* STEs with lookup type FLEX_PARSER_{0/1} includes
 	 * flex parsers_{0-3}/{4-7} respectively.
 	 */
-	sb->lu_type = sb->caps->flex_parser_id_mpls_over_udp > DR_STE_MAX_FLEX_0_ID ?
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1 :
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0;
+	sb->lu_type = sb->caps->flex_parser_id_mpls_over_udp <= DR_STE_MAX_FLEX_0_ID ?
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0 :
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1;
 
 	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
 	sb->ste_build_tag_func = &dr_ste_v0_build_tnl_mpls_over_udp_tag;
@@ -1280,9 +1337,9 @@ static void dr_ste_v0_build_tnl_mpls_over_gre_init(struct dr_ste_build *sb,
 	/* STEs with lookup type FLEX_PARSER_{0/1} includes
 	 * flex parsers_{0-3}/{4-7} respectively.
 	 */
-	sb->lu_type = sb->caps->flex_parser_id_mpls_over_gre > DR_STE_MAX_FLEX_0_ID ?
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1 :
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0;
+	sb->lu_type = sb->caps->flex_parser_id_mpls_over_gre <= DR_STE_MAX_FLEX_0_ID ?
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0 :
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1;
 
 	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
 	sb->ste_build_tag_func = &dr_ste_v0_build_tnl_mpls_over_gre_tag;
@@ -1347,9 +1404,9 @@ static void dr_ste_v0_build_icmp_init(struct dr_ste_build *sb,
 	is_ipv4 = DR_MASK_IS_ICMPV4_SET(&mask->misc3);
 	parser_id = is_ipv4 ? sb->caps->flex_parser_id_icmp_dw0 :
 			      sb->caps->flex_parser_id_icmpv6_dw0;
-	sb->lu_type = parser_id > DR_STE_MAX_FLEX_0_ID ?
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1 :
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0;
+	sb->lu_type = parser_id <= DR_STE_MAX_FLEX_0_ID ?
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0 :
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1;
 	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
 	sb->ste_build_tag_func = &dr_ste_v0_build_icmp_tag;
 }
@@ -1488,9 +1545,9 @@ dr_ste_v0_build_flex_parser_tnl_geneve_tlv_opt_init(struct dr_ste_build *sb,
 	/* STEs with lookup type FLEX_PARSER_{0/1} includes
 	 * flex parsers_{0-3}/{4-7} respectively.
 	 */
-	sb->lu_type = sb->caps->flex_parser_id_geneve_opt_0 > DR_STE_MAX_FLEX_0_ID ?
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1 :
-		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0;
+	sb->lu_type = sb->caps->flex_parser_id_geneve_opt_0 <= DR_STE_MAX_FLEX_0_ID ?
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_0 :
+		      DR_STE_V0_LU_TYPE_FLEX_PARSER_1;
 
 	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
 	sb->ste_build_tag_func = &dr_ste_v0_build_flex_parser_tnl_geneve_tlv_opt_tag;
@@ -1503,8 +1560,8 @@ static int dr_ste_v0_build_flex_parser_tnl_gtpu_tag(struct dr_match_param *value
 	struct dr_match_misc3 *misc3 = &value->misc3;
 
 	DR_STE_SET_TAG(flex_parser_tnl_gtpu, tag,
-		       gtpu_flags, misc3,
-		       gtpu_flags);
+		       gtpu_msg_flags, misc3,
+		       gtpu_msg_flags);
 	DR_STE_SET_TAG(flex_parser_tnl_gtpu, tag,
 		       gtpu_msg_type, misc3,
 		       gtpu_msg_type);
@@ -1525,6 +1582,60 @@ static void dr_ste_v0_build_flex_parser_tnl_gtpu_init(struct dr_ste_build *sb,
 	sb->ste_build_tag_func = &dr_ste_v0_build_flex_parser_tnl_gtpu_tag;
 }
 
+static int
+dr_ste_v0_build_tnl_gtpu_flex_parser_0_tag(struct dr_match_param *value,
+					   struct dr_ste_build *sb,
+					   uint8_t *tag)
+{
+	if (sb->caps->flex_parser_id_gtpu_dw_0 <= DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_dw_0, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_teid <= DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_teid, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_dw_2 <= DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_dw_2, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_first_ext_dw_0 <= DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_first_ext_dw_0, sb->caps, &value->misc3);
+	return 0;
+}
+
+static void
+dr_ste_v0_build_tnl_gtpu_flex_parser_0_init(struct dr_ste_build *sb,
+					    struct dr_match_param *mask)
+{
+	dr_ste_v0_build_tnl_gtpu_flex_parser_0_tag(mask, sb, sb->bit_mask);
+
+	sb->lu_type = DR_STE_V0_LU_TYPE_FLEX_PARSER_0;
+	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
+	sb->ste_build_tag_func = &dr_ste_v0_build_tnl_gtpu_flex_parser_0_tag;
+}
+
+static int
+dr_ste_v0_build_tnl_gtpu_flex_parser_1_tag(struct dr_match_param *value,
+					   struct dr_ste_build *sb,
+					   uint8_t *tag)
+{
+	if (sb->caps->flex_parser_id_gtpu_dw_0 > DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_dw_0, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_teid > DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_teid, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_dw_2 > DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_dw_2, sb->caps, &value->misc3);
+	if (sb->caps->flex_parser_id_gtpu_first_ext_dw_0 > DR_STE_MAX_FLEX_0_ID)
+		DR_STE_SET_FLEX_PARSER_FIELD(tag, gtpu_first_ext_dw_0, sb->caps, &value->misc3);
+	return 0;
+}
+
+static void
+dr_ste_v0_build_tnl_gtpu_flex_parser_1_init(struct dr_ste_build *sb,
+					    struct dr_match_param *mask)
+{
+	dr_ste_v0_build_tnl_gtpu_flex_parser_1_tag(mask, sb, sb->bit_mask);
+
+	sb->lu_type = DR_STE_V0_LU_TYPE_FLEX_PARSER_1;
+	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
+	sb->ste_build_tag_func = &dr_ste_v0_build_tnl_gtpu_flex_parser_1_tag;
+}
+
 static int dr_ste_v0_build_register_0_tag(struct dr_match_param *value,
 					  struct dr_ste_build *sb,
 					  uint8_t *tag)
@@ -1681,6 +1792,27 @@ static void dr_ste_v0_build_flex_parser_1_init(struct dr_ste_build *sb,
 	sb->ste_build_tag_func = &dr_ste_v0_build_flex_parser_tag;
 }
 
+static int dr_ste_v0_build_tunnel_header_0_1_tag(struct dr_match_param *value,
+						 struct dr_ste_build *sb,
+						 uint8_t *tag)
+{
+	struct dr_match_misc5 *misc5 = &value->misc5;
+
+	DR_STE_SET_TAG(tunnel_header, tag, tunnel_header_dw0, misc5, tunnel_header_0);
+	DR_STE_SET_TAG(tunnel_header, tag, tunnel_header_dw1, misc5, tunnel_header_1);
+
+	return 0;
+}
+
+static void dr_ste_v0_build_tunnel_header_0_1_init(struct dr_ste_build *sb,
+						   struct dr_match_param *mask)
+{
+	sb->lu_type = DR_STE_V0_LU_TYPE_TUNNEL_HEADER;
+	dr_ste_v0_build_tunnel_header_0_1_tag(mask, sb, sb->bit_mask);
+	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
+	sb->ste_build_tag_func = &dr_ste_v0_build_tunnel_header_0_1_tag;
+}
+
 static struct dr_ste_ctx ste_ctx_v0 = {
 	/* Builders */
 	.build_eth_l2_src_dst_init	= &dr_ste_v0_build_eth_l2_src_dst_init,
@@ -1703,11 +1835,14 @@ static struct dr_ste_ctx ste_ctx_v0 = {
 	.build_tnl_geneve_init		= &dr_ste_v0_build_flex_parser_tnl_geneve_init,
 	.build_tnl_geneve_tlv_opt_init	= &dr_ste_v0_build_flex_parser_tnl_geneve_tlv_opt_init,
 	.build_tnl_gtpu_init		= &dr_ste_v0_build_flex_parser_tnl_gtpu_init,
+	.build_tnl_gtpu_flex_parser_0	= &dr_ste_v0_build_tnl_gtpu_flex_parser_0_init,
+	.build_tnl_gtpu_flex_parser_1	= &dr_ste_v0_build_tnl_gtpu_flex_parser_1_init,
 	.build_register_0_init		= &dr_ste_v0_build_register_0_init,
 	.build_register_1_init		= &dr_ste_v0_build_register_1_init,
 	.build_src_gvmi_qpn_init	= &dr_ste_v0_build_src_gvmi_qpn_init,
 	.build_flex_parser_0_init	= &dr_ste_v0_build_flex_parser_0_init,
 	.build_flex_parser_1_init	= &dr_ste_v0_build_flex_parser_1_init,
+	.build_tunnel_header_0_1	= &dr_ste_v0_build_tunnel_header_0_1_init,
 	/* Getters and Setters */
 	.ste_init			= &dr_ste_v0_init,
 	.set_next_lu_type		= &dr_ste_v0_set_next_lu_type,
@@ -1717,14 +1852,16 @@ static struct dr_ste_ctx ste_ctx_v0 = {
 	.set_hit_addr			= &dr_ste_v0_set_hit_addr,
 	.set_byte_mask			= &dr_ste_v0_set_byte_mask,
 	.get_byte_mask			= &dr_ste_v0_get_byte_mask,
+	.set_ctrl_always_hit_htbl	= &dr_ste_v0_set_ctrl_always_hit_htbl,
+	.set_ctrl_always_miss		= &dr_ste_v0_set_ctrl_always_miss,
 	/* Actions */
+	.actions_caps			= DR_STE_CTX_ACTION_CAP_NONE,
 	.set_actions_rx			= &dr_ste_v0_set_actions_rx,
 	.set_actions_tx			= &dr_ste_v0_set_actions_tx,
-	.modify_field_arr_sz		= ARRAY_SIZE(dr_ste_v0_action_modify_field_arr),
-	.modify_field_arr		= dr_ste_v0_action_modify_field_arr,
 	.set_action_set			= &dr_ste_v0_set_action_set,
 	.set_action_add			= &dr_ste_v0_set_action_add,
 	.set_action_copy		= &dr_ste_v0_set_action_copy,
+	.get_action_hw_field		= &dr_ste_v0_get_action_hw_field,
 	.set_action_decap_l3_list	= &dr_ste_v0_set_action_decap_l3_list,
 };
 
